# These modules provide mixin examples that are aimed at boosting agent development speed.
# They also offer a foundation for creating your own custom mixins.

from .folium import Folium
from .langchain import LangChainTools
from .openai import OpenAIAPI
